import json
import random
import sys
import math

import numpy as np

from data_generation.generation import sample, Specification
from length_result import refineFrequencies as toDst
from patternLength import Pattern
from apply_pattern import applyPattern


def parseJSONandGenerate(file_path, apply_seed = True):
    with open(file_path) as json_file:
        specification = json.load(json_file)

    if apply_seed:
        random.seed(specification['random_seed'])

    X, Y, deltaSequ = genXY(specification)

    openY = np.where(Y == 0)[0]
    falseIE = random.sample(set(openY), int(specification['false_IE']))
    for i in falseIE:
        Y[i] = 1

    sampled_dst = {}
    for pattern, d_sequ in deltaSequ.items():
        sampled_dst[pattern] = toDst(d_sequ)


    spec = Specification(str(specification) + "Sampled Dst: " + str(sampled_dst), sampled_dst,
                         deltaSequ)
    return list(X), list(Y), spec




def check_open(pattern_locations,open_x):
    for loc in pattern_locations:
        if loc not in open_x:
            return False
    return True


def get_locations(pattern, planted_patterns):
    assert isinstance(pattern,tuple)
    locations = set()
    for p, l , d in planted_patterns.values():
        if p == pattern:
            locations.add(l)
    return locations


def fill_open_x(X, open_x, alphabet):
    for l in open_x:
        x = random.sample(alphabet, 1)[0]
        X[l] = x

def genXY(specification):
    length = specification['length']
    alphabet = specification['generating_alphabet']

    if 0 in alphabet or '0' in alphabet:
        raise ValueError('0 is not allowed in alphabet')

    open_y = set(range(length))
    open_x = set(range(length))
    padding_x = set()

    X = np.zeros(length, dtype=object)
    Y = np.zeros(length, dtype=int)

    all_planted_patterns = {}
    reverted_index = {}

    pattern_counters = {}
    pattern_distributions = {}

    for pattern_spec in specification['patterns']:
        pattern = pattern_spec['pattern']
        occurrences = pattern_spec['occurrences']
        dst = pattern_spec['distribution']
        gap_factor = pattern_spec['gap_factor']
        pattern_counters[tuple(pattern)] = occurrences
        pattern_distributions[tuple(pattern)] = dst


    pattern_id_counter = 0
    fill_open_x(X, open_x, alphabet)

    while sum(pattern_counters.values()) > 0:

        if (len(open_x) < 0.8 * len(X)):
            print("Only: {} free locations left".format(str(len(open_x)/len(X))), file=sys.stderr)

        next_pattern = random.sample(list(filter(lambda x: pattern_counters[x] != 0,pattern_counters.keys())),1)[0]

        if pattern_counters[next_pattern] == 0:
            continue

        delta = sample(pattern_distributions[next_pattern])

        pattern_loc = random.sample(open_x, 1)[0]

        pattern_locations = sorted(random.sample(range(pattern_loc,pattern_loc +math.ceil(len(next_pattern)*gap_factor)),len(next_pattern)))

        if check_open(pattern_locations, open_x) and (delta == 'skip' or (max(pattern_locations) + delta) in open_y):
            # write pattern into X and Y
            print('Plant Pattern: ', next_pattern, file=sys.stderr)

            for i,p in zip(pattern_locations, next_pattern):
                X[i] = p
                open_x.remove(i)
                reverted_index[i] = pattern_id_counter
                
            if delta != 'skip':
                ly = max(pattern_locations) + delta
                Y[ly] = 1
                open_y.remove(ly)
            else:
                delta = None

            all_planted_patterns[pattern_id_counter] = (next_pattern, max(pattern_locations), delta)
            pattern_id_counter += 1

            pattern_counters[next_pattern] -= 1
    delta_sequ = {}
    sorted_planted = sorted(all_planted_patterns.values(), key=lambda x:x[1])
    for p, i, d in sorted_planted:
        if p not in delta_sequ.keys():
            delta_sequ[p] = []
        delta_sequ[p].append((i,d))

    return X,Y, delta_sequ