import argparse
import json
import os
import random
import sys


def issubsequ(x,y, allow_gaps= False):
    if (allow_gaps):
        yIt = 0
        for i in range(len(x)):
            if yIt == len(y):
                return False;
            while(x[i] != y[yIt]):
                yIt += 1
                if yIt == len(y):
                    return False;
            yIt +=1
        return True
    else:
        if len(x) > len(y): return False

        ly = list(y)
        for i,yi in enumerate(ly):
            match = True
            for j, xy in enumerate(list(x)):
                if j+i >= len(ly) or xy != ly[j+i] :
                    match =  False
                    break
            if match:
                return True
        return False


def check_seen(p, seen_pattern):
    for seen in seen_pattern:
        if issubsequ(p,seen) or issubsequ(seen,p):
            return True
    return False


def sample_pattern(l,alphabet):
    pattern = []
    for _ in range(l):
        p = random.sample(alphabet, 1)[0]
        pattern.append(p)
    return pattern



def gen_patterns(args, distribution_spec):
    seen_patterns = set()
    predictive_occurences = 0.0

    patterns = []
    pattern_counter = args.predictive_patterns


    while pattern_counter > 0:
        p_length = random.randint(args.min_pattern_length,args.max_pattern_length)
        pattern = sample_pattern(p_length,alphabet)
        if not check_seen(tuple(pattern), seen_patterns):
            seen_patterns.add(tuple(pattern))
            pattern_counter -= 1
            pattern_spec = {}
            pattern_spec['pattern'] = pattern
            pattern_spec['gap_factor'] = args.gap_factor
            occurences = random.randint(args.min_occurences, args.max_occurences)
            pattern_spec['occurrences'] = round(occurences)
            predictive_occurences += (1-distribution_spec['skip']) * occurences
            pattern_spec['distribution'] = distribution_spec
            patterns.append(pattern_spec)

    frequ_counter = args.frequent_patterns

    while frequ_counter > 0:
        p_length = random.randint(args.min_pattern_length, args.max_pattern_length)
        pattern = sample_pattern(p_length,alphabet)
        if not check_seen(tuple(pattern), seen_patterns):
            seen_patterns.add(tuple(pattern))
            frequ_counter -= 1
            pattern_counter -= 1 #schould not be here but I'm afraid of deleting it :)
            pattern_spec = {}
            pattern_spec['pattern'] = pattern
            pattern_spec['gap_factor'] = args.gap_factor
            occurences = random.randint(args.min_occurences, args.max_occurences)
            pattern_spec['occurrences'] = round(occurences)
            pattern_spec['distribution'] = {'skip':1, 'peaks':[], 'peak_distribution':[]}
            patterns.append(pattern_spec)

    return patterns, predictive_occurences




def dst_spec(args):
    distribution_spec = {}
    distribution_spec['skip'] = args.skip_prob
    dst_args = args.dst_args
    arg0 = dst_args.split(',')[0].split(':')
    arg1 = dst_args.split(',')[1].split(':')


    peaks = [{'name': args.distribution, arg0[0]:int(arg0[1]), arg1[0]:int(arg1[1])}]
    peak_dst = [1]

    distribution_spec['peaks'] = peaks
    distribution_spec['peak_distribution'] = peak_dst

    return distribution_spec



if __name__ == '__main__':


    parser = argparse.ArgumentParser()

    '''
    Example Usage:
    -experiments
1
-sequ_length
50000
-min_occurences
4000
-max_occurences
5000
-predictive_patterns
1
-frequent_patterns
0
-max_pattern_length
6
-min_pattern_length
5
-distribution
normal
-dst_args
median:10,std_deviation:2
-false_ie
0
-skip_prob
0
-alphabet_size
50
-save_to
./large_lag
--override
    '''

    parser.add_argument('-experiments', type=int)
    parser.add_argument('-sequ_length', type=int)
    parser.add_argument('-min_occurences', type=int)
    parser.add_argument('-max_occurences', type=int)
    parser.add_argument('-predictive_patterns', type=int)
    parser.add_argument('-frequent_patterns', type=int)
    parser.add_argument('-max_pattern_length', type=int)
    parser.add_argument('-min_pattern_length', type=int)
    parser.add_argument('-gap_factor', type=float)
    parser.add_argument('-distribution',  type=str)
    parser.add_argument('-dst_args', type=str)
    parser.add_argument('-false_ie', type=float)
    parser.add_argument('-skip_prob',type=float)
    parser.add_argument('-alphabet_size', type=int)
    parser.add_argument('-save_to', type=str)
    parser.add_argument('--override', default=False, action='store_true')
    parser.add_argument('-random_seed', type=str, default='omen')

    args = parser.parse_args()

    random.seed(args.random_seed);

    if os.path.isdir(args.save_to):
        if not args.override:
            raise FileExistsError('Folder already exists!!')
    else:
        os.mkdir(args.save_to)

    alphabet = list(range(1,args.alphabet_size+1))
    distribution_spec = dst_spec(args)

    for e in range(args.experiments):
        e_spec = {}
        seed = random.randrange(sys.maxsize)
        rng = random.Random(seed)
        e_spec['random_seed'] = seed
        e_spec['generating_alphabet'] = alphabet
        e_spec['length'] = args.sequ_length

        patterns, predictive_occurences  = gen_patterns(args,distribution_spec)

        p = args.false_ie
        e_spec['false_IE'] = round(p * predictive_occurences / (1-p))


        e_spec['patterns'] = patterns


        e_file = os.path.join(args.save_to,'experiment_{}.json'.format(e))
        f = open(e_file, 'w+')
        json.dump(e_spec,f,indent=4)
        f.flush()
        f.close()
