import json
import random

import numpy as np


class Specification:
    def __init__(self, out_str, planted_dst = None, sampled_delta_sequ = None):
        self.out_str = out_str
        self.planted_dst = planted_dst
        self.sampled_delta_sequ = sampled_delta_sequ

def genreateX(length, alphabet):
    X = []
    for _ in range(length):
        X.append(random.sample(alphabet,1)[0])
    return X

def sample(dst_specification):
    if random.random() < dst_specification['skip']:
        return 'skip'
    else:
        peak_dst = dst_specification['peak_distribution']
        norm_peak_dst = list(map(lambda x: x / sum(peak_dst), peak_dst))
        peak_sample = random.random()
        peak_border = 0
        peaks = dst_specification['peaks']
        for peak_i, peak_prob in enumerate(norm_peak_dst):
            peak_border += peak_prob
            if peak_sample <= peak_border:
               return sampleFromPeak(peaks[peak_i])

        # just as a catch all if something goes wrong
        return sampleFromPeak(peaks[len(peaks)-1])




def sampleFromPeak(peak_specification):
    name = peak_specification['name']

    if name == 'uniform':
        return sampleUniform(peak_specification)
    elif name == 'normal':
        return sampleNormal(peak_specification)
    else:
        raise ValueError('Invalid peaks specification ' + str(peak_specification))



def sampleUniform(peak_specification):
    max_delta = peak_specification['max_delta']
    min_delta = peak_specification['min_delta']
    sample = random.randint(min_delta, max_delta)
    return sample

def sampleNormal(peak_specification):
    median = peak_specification['median']
    std_deviation = peak_specification['std_deviation']
    sample = round(random.gauss(median, std_deviation))
    return sample



def parseJSONandGenerate(file_path, apply_seed = True):
    with open(file_path) as json_file:
        specification = json.load(json_file)

    if apply_seed:
        random.seed(specification['random_seed'])

    length = specification['length']
    X = genreateX(length, specification['generating_alphabet'])
    Y = np.zeros(length, dtype=int)

    sampled_dst = {}

    for pattern_specification in specification['patterns']:
        pattern = pattern_specification['pattern']
        pattern_key = tuple(pattern)
        sampled_dst[pattern_key] = {}
        occurrences = pattern_specification['occurrences']
        pattern_locations = random.sample(range(length),occurrences)
        for pattern_loc in pattern_locations:
            try:
                for i, p_i in enumerate(pattern):
                    X[pattern_loc +i] = p_i
                ie_loc = sample(pattern_specification['distribution'])
                if ie_loc in sampled_dst[pattern_key]:
                    sampled_dst[pattern_key][ie_loc] += 1
                else:
                    sampled_dst[pattern_key][ie_loc] = 1
                if ie_loc != 'skip':
                    Y[pattern_loc+len(pattern)+ie_loc -1] = 1
            except IndexError:
                pass
    spec = Specification(str(specification) + "  Sampled Dst: " + str(sampled_dst), sampled_dst)
    return X,list(Y), spec