import copy
import math

import numpy as np

import length_result as lm2
from dst_length_helper import universal_interger_encoding
import dst_length


class Pattern:

    def __init__(self, pattern):
        self.pattern = pattern

    def extend(self,element, return_copy = False):
        if return_copy:
            copy_of_sequ = copy.deepcopy(self.pattern)
            copy_of_sequ.append(element)
            return Pattern(copy_of_sequ)
        else:
            self.pattern.append(element)

    def extendFront(self, element, return_copy = False):
        if return_copy:
            copy_of_sequ = copy.deepcopy(self.pattern)
            new_sequ = [element] + copy_of_sequ
            return Pattern(new_sequ)
        else:
            self.pattern = [element] + self.pattern

    def addAlignment(self, deltaSequ):
        self.deltaSequ = deltaSequ
        self.frequencies = lm2.refineFrequencies(deltaSequ)

    def getPatternName(self):
        return str(self.pattern)

    def __str__(self):
        if hasattr(self,'deltaSequ'):
            return "Pattern: {} Frequencies: {} Delta-Sequ: {}". format(str(self.pattern), str(self.frequencies) , str(self.deltaSequ))
        else:
            return str(self.pattern)
    def __repr__(self):
        return self.__str__()


def build_frequ_over_X(X):
    X = np.array(X, dtype=object)
    frequencies = {}
    unique_items  = np.unique(X)
    for item in unique_items:
        frequencies[item] = 1
    return frequencies

def compute_length(pattern, frequencies):
    assert isinstance(pattern, Pattern)
    length_of_X = sum(frequencies.values())
    length = 0
    for element in pattern.pattern:
        frequencie = frequencies[element]
        length -= math.log(frequencie/length_of_X,2)
    return length

def compute_length_alphabet(pattern,alphabet):
    assert isinstance(pattern, Pattern)
    n_symbols = universal_interger_encoding(len(pattern.pattern), dst_length.C)
    length = len(pattern.pattern) * (-math.log(1/len(alphabet),2))
    return length + n_symbols
